/*
 * botactivity.h
 *
 *  Created on: 17 Sep 2015
 *      Author: irina
 */

#ifndef BOTACTIVITY_H_
#define BOTACTIVITY_H_

#include "net/message.h"
#include "botlog.h"
#include <csutil/ref.h>
#include <csutil/set.h>
#include <csutil/bitarray.h>

#include "worldhandler.h"
#include "botstorage.h"
#include "recipe.h"

#include "gui/pawsslot.h"


//generic states for any bot activity
enum ACTIVITY_STATES
{
	ERROR_ = -1,
	ONGOING_ = 0,
	FINISHED_ = 1,
	STOPPED_ = 2,
	READY_
};
//explore strategies for the explore activity
enum EXPLORE_STRATEGIES
{
	EXPLORE_RADIAL = 1,
	EXPLORE_GRID = 2,
	EXPLORE_LINE = 3
};

//step-wise actions done by the crafting activity between imposed/forced delays
enum CRAFT_ACTIONS
{
	//CRAFTING_LEARNSTORAGE = 21,
	CRAFTING_INIT = 21,
	CRAFTING_READRECIPE = 22,
	//CRAFTING_GETINGREDIENTS = 23,
	CRAFTING_USE_COMBINE = 24,
	CRAFTING_TAKE_RESULT = 25,
	CRAFTING_TRAIN = 26,
	CRAFTING_NONE
};

//step-wise actions done by the explore activity between imposed/forced delays
enum EXPLORE_ACTIONS
{
	ACTION_INIT = 10,
	ACTION_DROP_TABLE = 11,
	ACTION_SCAN_TABLE = 12,
	ACTION_TEST_TABLE = 13,
	ACTION_EXPLORE = 14,
	ACTION_READRECIPE = 15,
	ACTION_BUILD = 16,
	ACTION_USE_COMBINE = 17,
	ACTION_EXAMINE_RESULT = 18,
	ACTION_TAKE_RESULT = 19,
	ACTION_DROP_KEY = 20,
	ACTION_STORE_RESOURCES = 21,
	ACTION_MOVE = 22,
	ACTION_UNMOVE = 23,
	ACTION_MOVE_TOWARD_TABLE = 24,
	ACTION_STOP = 25,
	ACTION_DIE = 26,
	ACTION_NONE
};

enum REPAIR_ACTIONS
{
	REPAIR_INIT = 20,
	REPAIR_WORK = 21,
	REPAIR_TRAIN = 22,
	REPAIR_NONE = 23
};

const int MAX_DELAY = 8000;
const int MIN_DELAY = 600;
const int DEATH_DELAY = 62000;
const int DEFAULT_SIZE = 3;
const int DEFAULT_THREADS_SMALL_CLAIM = 7;
const int ITEM_COUNT_TINY_CLAIMS = 1;
const int TEXT_COLOUR = 61046;
const int TEXT_COLOUR_WHITE = 0xeeeeee;
const int MAX_ATTEMPTS = 4;
const int DEFAULT_EXPLORE_TIMEOUT = 30000;


const int ANGLE_RADIAL_STRATEGY = 5;
const EXPLORE_STRATEGIES DEFAULT_STRATEGY = EXPLORE_RADIAL;
const csVector3 MIDDLE_MAP_POS = csVector3(0, 52.26f, 0);
const csVector3 TRAINER_POS = csVector3(174, 57, 159);
static const csString EXP_MARKER = csString("Exploration Marker");
static const csString NOT_SUCCESSFUL = csString("You were not successful");
static const csString UNSUCCESSFUL_SINCE_MOVED = csString("You were unsuccessful since you moved");
static const csString UNSUCCESSFUL_SINCE_CHANGED = csString("You were unsuccessful since you changed something");
static const csString CANT_MOVE_ITEMS = csString("You can't move items to that");
static const csString EXPLORE_TOOL_WORN = csString("That tool is not useful");
static const csString CLAIM_PLACED = csString("You placed an ");
static const csString GOT_ITEMS = csString("You got a ");
static const csString ITEM_KEY = csString("Key");
static const csString ITEM_ENUM = csString("Ennumeration");
static const csString DIEING = csString("Your hands are raw, your knees buckle");
static const csString OVERWEIGHT = csString("You struggle under the weight of your inventory and must drop something");
static const csString RANKING_PREFIX = csString("You've ranked up in ");
static const csString CONTAINER_WORN = csString("The container is too worn to be useful");
static const csString TOOL_WORN = csString("You do not have a good tool");
static const csString NOT_IN_RANGE = csString("You are not in range to ");
static const csString FINISHED_WORKING = csString("You finished working");
static const csString TOO_FAR_AWAY_TO_USE = csString("You are too far away to use that");
static const csString IT_DOESNT_FIT_THERE = csString("It doesn't fit there");
static const csString MAXXED_SKILL = csString("You practiced");
static const csString CANNOT_EXPLORE_TWICE = csString("You cannot explore in the same place twice");
static const csString CRAFT_TABLE = csString("Craft-Table");
static const csString LOOT_ON_GROUND = csString("You've got loot but some of it fell on the ground");


static const csString TOO_FAR_AWAY_TO_PICKUP = csString("That item is too far away to pick up");
static const csString CANNOT_BE_PICKED_UP = csString("That item cannot be picked up");
static const csString TOO_FAT_TO_PICKUP = csString("You are either too fat or too busy to start /pickup");
static const csString YOU_DROPPED = csString("You dropped");

class slot
{
public:
	slot() { Clear(); }
	slot(int s, csString n, int64_t c, int q, bool inv) : slotNo(s), name(n), count(c), quality(q), isInInventory(inv) {}

	void Clear(int s = -1)
	{
		slotNo = s;
		name = csString("");
		count = 0;
		quality = 0;
		container = 0;
		isInInventory = false;
	}

	int slotNo;
	csString name;
	int64_t count;
	int quality;
	int container;
	bool isInInventory;
};

inline bool operator< (const slot& lhs, const slot& rhs){ 
	if (lhs.quality == rhs.quality)
		return rhs.isInInventory;
	return (lhs.quality < rhs.quality); 
}
inline bool operator> (const slot& lhs, const slot& rhs){ return rhs < lhs; }
inline bool operator<=(const slot& lhs, const slot& rhs){ return !(lhs > rhs); }
inline bool operator>=(const slot& lhs, const slot& rhs){ return !(lhs < rhs); }
inline bool operator==(const slot& lhs, const slot& rhs){ 
	return lhs.slotNo == rhs.slotNo && lhs.quality == rhs.quality && lhs.isInInventory == rhs.isInInventory; 
}

class slotpair 
{
public:
	slot a;
	slot b;

	slotpair(slot s1, slot s2) : a(s1), b(s2) {}

	int64_t minCount() const {
		return a > b ? b.count : a.count;
	}
};

// sort order is first by quality difference decending (i.e. reverse), then by min count ascending
inline bool operator< (const slotpair& lhs, const slotpair& rhs) {
	int lhsDiff = abs(lhs.a.quality - lhs.b.quality);
	int rhsDiff = abs(rhs.a.quality - rhs.b.quality);
	if (lhsDiff != rhsDiff)
		return lhsDiff > rhsDiff; // reverse
	return lhs.minCount() < rhs.minCount();
}
inline bool operator> (const slotpair& lhs, const slotpair& rhs){ return rhs < lhs; }
inline bool operator<=(const slotpair& lhs, const slotpair& rhs){ return !(lhs > rhs); }
inline bool operator>=(const slotpair& lhs, const slotpair& rhs){ return !(lhs < rhs); }

class move
{
public:
	slot src;
	slot dest;
	int64_t count;

	move() : src(slot()), dest(slot()), count(0) {}
	move(slot s1, slot s2, int64_t c) : src(s1), dest(s2), count(c) {}
};


//generic class for bot activities - to create a new activity, inherit from this class
class BotActivity
{
public:

	virtual void HandleMessage(MsgEntry* message ) =0;

	csString GetActivityName() {
		return name;
	}

	void OutputMsg(csString msg)
	{
		OutputMsg(msg, textcolour);
	}
	void OutputMsg(csString msg, int color)
	{
		if (textbox != NULL)
		{
			csString out = csString(">") + msg;
			textbox->Show();
			textbox->AddMessage(out.GetData(), color);
		}
	}

	void Init(BotLogCSV* logger, int timesToRepeat, pawsMessageTextBox* textbox, csString cmd) {
		this->logger = logger;
		this->timesToRepeat = timesToRepeat;
		this->textbox = textbox;
		timesDone = 0;
		startPoint = worldHandler::GetCurrentPos();
		startYRot = worldHandler::GetCurrentRot();
		this->cmd = cmd;

		
        state = READY_;
		DoInit();
	}
	void Start() {
		state = ONGOING_;
		StartWork();
	}
	void Stop() {
		state = STOPPED_;
		WrapUp();
	}
	void Finish() {
		timesDone = timesDone + 1;
		WrapUp();
		state = FINISHED_;
	}
	void Restart() {
		Stop();
		Start();
	}
	void Error() {
		state = ERROR_;
		WrapUp();
	}

	void Reset() {
		Stop();
		Init(logger, timesToRepeat, textbox, cmd);
	}

	bool IsOngoing() {
		return state == ONGOING_ || (state == FINISHED_ && timesDone < timesToRepeat);
	}

	bool IsError() {
		return state == ERROR_;
	}

	bool IsFinished() {
		return state == FINISHED_;
	}

	bool IsStopped() {
		return state == STOPPED_;
	}
	bool IsReady() {
		return state == READY_;
	}
	bool Compare(csVector3 a, csVector3 b) {
		float epsilon = 0.01f;
		float difx = fabs(a.x - b.x);
		float dify = fabs(a.y - b.y);
		float difz = fabs(a.z - b.z);
		printf("comparing %s, %s)\n", a.Description().GetData(), b.Description().GetData());
		printf("    diffs: %f, %f, %f\n", difx, dify, difz);
		return difx < epsilon && dify < epsilon && difz < epsilon;
	}

	virtual void ScheduleNewPerform(int delay) =0;

private:
	ACTIVITY_STATES state;

protected:
	csString name;		//name of activity, intended to be set by classes inheriting this one

    pawsMessageTextBox* textbox;
    int textcolour;
    csString cmd;

	int timesToRepeat;
	int timesDone;
	csTicks timeout, startTime;

	csVector3 startPoint;
	float startYRot;
	csString skillToTrain;

	virtual void StartWork() =0;
	virtual void WrapUp() =0;
	virtual void DoInit() =0;

	BotLogCSV* logger;
};


class MovementStrategy
{
public:
	MovementStrategy();
	void Move(int movesDone);
	void UnMove(int movesDone);
	void Init(EXPLORE_STRATEGIES strategy, int size, csVector3 startPos, float yRot)
	{
		this->strategy = strategy;
		this->size = size;
		this->startPos = startPos;
		this->startYRot = yRot;
	}

protected:
	EXPLORE_STRATEGIES strategy;
	int size;						//basically a measure of the area to cover

	csVector3 startPos;
	float startYRot;
private:
	csVector3 currPos;
};

class ExploreActivity: public BotActivity, iTimerEvent
{
public:
	ExploreActivity();
	~ExploreActivity();
	void HandleMessage(MsgEntry* message );
	void ScheduleNewPerform(int delay);
	bool Perform(iTimerEvent *ev);

	//void SetStrategy(MovementStrategy ms);
	MovementStrategy GetStrategy() {
		return strategy;
	}

	//for iBase
			void IncRef(){}
			void DecRef(){}
			int GetRefCount() {return 0;}
			void* QueryInterface(scfInterfaceID i, int a) {return NULL;}
			void AddRefOwner(void **, CS::Threading::Mutex*) {}
			void RemoveRefOwner(void**){}
			scfInterfaceMetadataList* GetInterfaceMetadata() {return NULL;}

protected:

	void StartWork();
	void WrapUp();
	void DoInit();

	void NextAction(int delay);
	//action methods

	void DoDropTable();
	void DoScanTable();
	void DoTestTable();
	void DoExplore();
	void DoReadRecipe();
	bool DoBuild();
	void DoUseCombine();
	void DoExamineResult();
	void DoTakeResult();
	void DoDie();
	void DoDropKey();
	void DoStoreResources();
	void DoMove();
	void DoUnMove();
	void DoMoveTowardTable();
	void DoStop();

	bool PickupTable();
	int FindTableSlot();
	int CountTables();
	int SelectToolSlot(csString tool, int quality);
	int SelectIngredientSlot(csString name, int quantity);
	csBitArray InventorySnapshot();
	void DoSendTableTest();
	csArray<slot> FindTableSlots(csString name, int quality, int band = 1);
	void ProcessNextResourceMove();
	csArray<move> Stack(csArray<slot> newSlots);
	csArray<slot> SimulateMove(move nextMove, csArray<slot> tableSlots);
	move ComputeNextMove(csArray<slot> slots);
	move BuildMove(slot a, slot b);
	csArray<slotpair> Combinations(csArray<slot> slots);
	void printMove(move m);
	void printSlots(csArray<slot> slots);

private:
	int attemptsCount;
	csString claimName;
	csString scrollName;
	csString keyName;
	EID markerID;
	EID tableEid;
	csString claimType;
	csString itemName;
	int initialTableCount;
	uint32 preResultInvVersion;
	csBitArray invSnapshot;
	int qualityBand;
	int markerWaitCount;

	//a craft-table used to manage different qualities depending on claim type
	int tableSlot;
	psInventoryCache* exploreInvCache;

	csArray<int> tableScanSlots;
	slot tableContents[16];
	csList<move> pendingResourceMoves;
	move currentResourceMove;
	csHash<psPersistItem, EID> possibleClaims;
	csHash<EID, EID> lootBags;
	csHash<int64_t, csString> claimOutput;
	csSet<int> emptyTableSlots;
	csTicks tableTestStart, tableTestTimeout, lastVerifyTime;

	int maxDelay;
	csVector3 markerPos;
	csVector3 tablePos;
	float lastKnownTableDist;
	float prevRotation;

	//flags
	bool needToDie;
	bool noBuild;
	bool leaveKeys;
	bool exploreFinished;
	bool usingTable;	//whether a table has been found in inventory and is being used for managing resources of different qualities
	bool pickupDone;	//for using Craft-tables to manage different qualities AND overweight
	bool useMax;
	bool buildFinished;
	bool needTableScan;
	bool tableScanSent;
	bool needTableTest;
	bool tableTestSent;
	bool tableTestPassed;
	bool seekingTableUrgently;
	bool glitchPickup;
	bool parseClaimKey;
	bool verifyClaimTransfer;
	bool claimTransferSuccess;
	bool waitingForDrop;
	bool watchingInvForNewResources;
	bool tableScanInProgress;
	bool storingResourcesInProgress;
	bool storingResourcesComplete;
	bool verifyInventoryMove;
	bool verifyTableSlotMove;
	bool shouldBeStandingStill;
	bool waitForScroll;
	bool rebuild;
	bool pickupGroundLoot;

	MovementStrategy strategy;

	pawsSlot* nextFreeSlot;		//for getting rid of tiny+small keys
	pawsSlot* toolSlot;		//for worn-out tools

	recipe claimRcp;
	EXPLORE_ACTIONS actionToDo;
};

class CraftActivity: public BotActivity, iTimerEvent
{
public:
	CraftActivity();
	~CraftActivity();
	void HandleMessage(MsgEntry* message );
	void ScheduleNewPerform(int delay);
	bool Perform(iTimerEvent *ev);

	//for iBase
			void IncRef(){}
			void DecRef(){}
			int GetRefCount() {return 0;}
			void* QueryInterface(scfInterfaceID i, int a) {return NULL;}
			void AddRefOwner(void **, CS::Threading::Mutex*) {}
			void RemoveRefOwner(void**){}
			scfInterfaceMetadataList* GetInterfaceMetadata() {return NULL;}

protected:

	void StartWork();
	void WrapUp();
	void DoInit();

	void NextAction(int delay);
	void MoveToContainer();

	//actions methods
	void DoReadRecipe();
	void DoCombine();
	void DoTakeResult();
	void DoTrain();

private:
	bool needToTrain;
	int craftDelay;
	bool useMax;
	bool craftFinished;
	csString logText;
	int trainSubstage;

	EID containerEID;
	recipe rcp;

	CRAFT_ACTIONS actionToDo;
};

class RepairActivity: public BotActivity, iTimerEvent
{
public:
	RepairActivity();
	~RepairActivity();
	void HandleMessage(MsgEntry* message );
	void ScheduleNewPerform(int delay);
	bool Perform(iTimerEvent *ev);

	//for iBase
			void IncRef(){}
			void DecRef(){}
			int GetRefCount() {return 0;}
			void* QueryInterface(scfInterfaceID i, int a) {return NULL;}
			void AddRefOwner(void **, CS::Threading::Mutex*) {}
			void RemoveRefOwner(void**){}
			scfInterfaceMetadataList* GetInterfaceMetadata() {return NULL;}

protected:

	void StartWork();
	void WrapUp();
	void DoInit();

	void NextAction(int delay);

	//actions methods
	void DoRepairWork();
	void DoTrain();

private:
	bool needToTrain;
	int craftDelay;
	bool repairFinished;
	csString logText;

	pawsSlot* toolSlot;		//for worn-out tools

	EID containerEID;
	REPAIR_ACTIONS actionToDo;
};



#endif /* BOTACTIVITY_H_ */
