#include "botlog.h"
#include "util/log.h"

BotLogCSV::BotLogCSV(iConfigManager* configmanager, iVFS* vfs)
{
    std::pair<csString, const char*> logs[MAX_BOTCSV];
    size_t maxSize = 10*1024*1024;

    logs[CSV_RECIPES] = std::make_pair(csString("logrecipe.csv"), RECIPE_HEADER);
    logs[CSV_EXPLORING] = std::make_pair(csString("logexploring.csv"), EXPLORING_HEADER);
    logs[CSV_CRAFTING] = std::make_pair(csString("logcrafting.csv"), CRAFTING_HEADER);
    logs[CSV_POS] = std::make_pair(csString("logpos.csv"), POS_HEADER);

    for(int i = 0;i < MAX_BOTCSV;i++)
    {
        StartLog(logs[i].first, vfs, logs[i].second, maxSize, csvFile[i]);
    }
}

void BotLogCSV::StartLog(const char* logfile, iVFS* vfs, csString header, size_t maxSize, csRef<iFile>& csvFile)
{
	csString filename;
	filename.Format("/planeshift/userdata/logs/%s_%s",
	        	                        psengine->GetMainPlayerName(),
	        	                        logfile);
	filename.ReplaceAll(" ", "_");

        bool writeHeader;
        if (!vfs->Exists(filename))
        {

            csvFile = vfs->Open(filename,VFS_FILE_WRITE);
            writeHeader = true;
        }
        else
        {
            csvFile = vfs->Open(filename,VFS_FILE_APPEND);
            writeHeader = false;
        }

        if(csvFile.IsValid() && writeHeader)
        {
            csvFile->Write(header.GetData(), header.Length());
            csvFile->Flush();
        }
}

void BotLogCSV::Write(int type, csString& text)
{
    if (!csvFile[type])
        return;


    time_t curtime = time(NULL);
    struct tm *loctime;
    loctime = localtime (&curtime);
    csString buf(asctime(loctime));
    buf.Truncate(buf.Length()-1);

    buf.Append(", ");
    buf.Append(text);
    buf.Append("\n");

    csvFile[type]->Write(buf, buf.Length());
    csvFile[type]->Flush();

}
