/*
 * recipe.cpp
 *
 *  Created on: 15 Jul 2015
 *      Author: Diana Coman
 */

#include <psconfig.h>

#include "recipe.h"
#include "util/strutil.h"
#include "util/log.h"


recipe::recipe()
{
	itemName = csString("");
	bundleName = csString("");
	recipeName = csString("");
	containerName = csString("");
	toolName = csString("");

	ingredientsList = new csHash<int, csString>();
}

recipe::~recipe()
{
	delete ingredientsList;
}

bool recipe::HasTool()
{
	return !toolName.CompareNoCase("");
}

int recipe::GetQuantity(csString ingredient)
{
	return ingredientsList->Get(ingredient, 0);
}

void recipe::GetCountAndName(csString singleIngredient, bool useMax)
{
	csString separator(" and ");
	csString between("between ");
	csString space(" ");
	size_t pos, nextPos, length;
	csString ingredientName;
	int ingredientCount;
	csString currText;

	pos = singleIngredient.Find(between);
	if (pos < singleIngredient.Length())
	{
		if (useMax)
		{
			nextPos = singleIngredient.Find(separator);
			nextPos = nextPos + separator.Length();
			length = singleIngredient.Length() - nextPos;

			currText = singleIngredient.Slice(nextPos, length);
			nextPos = currText.Find(space);

			ingredientCount = atoi(currText.GetData());
			length = currText.Length() - nextPos - 2;
			ingredientName = currText.Slice(nextPos+1, length);
		}
		else
		{
			nextPos = singleIngredient.Find(separator);
			pos = pos + between.Length();
			length = nextPos - pos;
			currText = singleIngredient.Slice(pos, length);
			ingredientCount = atoi(currText.GetData());

			nextPos = nextPos+separator.Length();
			length = singleIngredient.Length() - nextPos;
			currText = singleIngredient.Slice(nextPos, length);

			pos = currText.Find(space);
			length = currText.Length() - pos-2;
			ingredientName = currText.Slice(pos+1, length);
		}
	}
	else
	{//"precise" recipe
		ingredientCount = atoi(singleIngredient.GetData());
		if (ingredientCount == 0)
		{
			//no ingredient count specified, so everything is ingredient name
			ingredientCount = 1;
			ingredientName = singleIngredient;
		}
		else
		{
			pos = singleIngredient.Find(" ") + 1;
			length = singleIngredient.Length()-pos;
			ingredientName = singleIngredient.Slice(pos, length);
		}
	}

//	Notify3(LOG_USER, "Ingredient=#%s# Count=%d", ingredientName.GetData(), ingredientCount);
	ingredientsList->Put(ingredientName, ingredientCount);
}

void recipe::ParseIngredients(csString ingredients, bool useMax)
{
	csString separator(",");
	csString between("between ");
	size_t pos, nextPos, length;
	csString ingredientName;
	int ingredientCount;
	csString currText;

	pos = 0;

	while (pos < ingredients.Length())
	{
		while (ingredients.GetAt(pos) == ' ') pos++;   //skip leading spaces if any

		nextPos = NextPosInString(ingredients, separator, pos);
		length = nextPos - pos;
		currText = ingredients.Slice(pos, length);

		GetCountAndName(currText, useMax);
		pos = nextPos+1;
	}

}

int recipe::NextPosInString(csString str, csString separator, int startPos)
{
	if (startPos >= str.Length())
		return str.Length();

	int nextPos = str.Find(separator, startPos);
	if (nextPos > str.Length())
		nextPos = str.Length();

	return nextPos;
}

csString recipe::GetTextBetweenWords(csString str, csString wordBefore, csString wordAfter)
{
	size_t pos1 = str.Find(wordBefore);
	if (pos1 > str.Length()) return csString("");

	size_t pos2 = str.Find(wordAfter, pos1);
	if (pos1 > str.Length()) return csString("");

	size_t start = pos1 + wordBefore.Length();
	size_t length = pos2 - start;

	return str.Slice(start, length);
}

void recipe::ParseRecipe(csString recipeText, bool useMax)
{
	ingredientsList->Empty();

	csString ingredients = GetTextBetweenWords(recipeText, csString("Combine "), csString(", into "));
	ParseIngredients(ingredients, useMax);

	bundleName = GetTextBetweenWords(recipeText, csString(", into "), csString("."));

	csString beforeWord = bundleName + csString(" into");
	itemName = GetTextBetweenWords(recipeText, beforeWord, csString(" using"));

	csString lastPart = GetTextBetweenWords(recipeText, csString("using"), csString("."));

	size_t pos = lastPart.Find("with");
	if (pos>lastPart.Length())
	{
		toolName = "";
		pos = lastPart.Find(".");
		containerName = lastPart.Slice(1, pos);
	}
	else
	{
		containerName = lastPart.Slice(1, pos-1);
		toolName = GetTextBetweenWords(lastPart, csString("with a "), csString("."));
	}

	itemName = itemName.Collapse();
	bundleName = bundleName.Collapse();
	containerName = containerName.Collapse();
	toolName = toolName.Collapse();

}
